# Copyright 2016 Adam Beckmeyer
#
# This file is part of Gyr.
#
# Gyr is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gyr is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along
# with Gyr.  If not, see <http://www.gnu.org/licenses/>.

import falcon
from . import resources
from .api import MatrixASHttpAPI


class Application(falcon.API):

    def __init__(self, hs_address, hs_token, *args, **kwargs):
        # This runs the __init__ method of falcon.API
        super().__init__(*args, **kwargs)
        self.hs_address = hs_address
        self.hs_token = hs_token

    def add_handlers(self, room_handler=None, transaction_handler=None,
                     user_handler=None):
        """Adds routes to Application that use specified handlers."""
        # Add all the normal matrix API routes
        if room_handler:
            room = resources.Room(room_handler,
                                  self.create_api)
            self.add_route("/rooms/{room_alias}", room)
        if transaction_handler:
            transaction = resources.Transaction(transaction_handler,
                                                self.create_api)
            self.add_route("/transactions/{txn_id}", transaction)
        if user_handler:
            user = resources.User(user_handler,
                                  self.create_api)
            self.add_route("/users/{user_id}", user)

    def create_api(self, identity=None):
        """Returns an instance of MatrixASHttpAPI."""
        return MatrixASHttpAPI(self.hs_address,
                               identity=identity,
                               token=self.hs_token)



# Copyright 2016 Adam Beckmeyer
#
# This file is part of Gyr.
#
# Gyr is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Gyr is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Gyr.  If not, see
# <http://www.gnu.org/licenses/>.

from . import utils


class Event:
    """Represents a matrix event sent by the HS.

    Usage:
        event = Event(json, api_factory)
    """

    def __init__(self, json, api_factory):
        """Instantiates Event instance.

        Args:
            json(dict): Event json from homeserver.
            api_factory(func): Creates api for calling homeserver.
        """
        self.json = json
        self.api_factory = api_factory

    @property
    def user(self):
        """Creates a User object when requested."""
        try:
            return self._user
        except AttributeError:
            if "sender" in self.json:
                mxid = self.json["sender"]
            else:
                mxid = self.json["user_id"]
            self._user = MatrixUser(mxid, self.api_factory(mxid))
            return self._user

    @property
    def room(self):
        """Creates a Room object when requested."""
        try:
            return self._room
        except AttributeError:
            room_id = self.json["room_id"]
            self._room = MatrixRoom(room_id, self.api_factory())
            return self._room

    @property
    def type(self):
        """Returns the type of event."""
        return self.json["type"]

    @property
    def content(self):
        """Returns content of event."""
        return self.json["content"]

    @property
    def timestamp(self):
        """Returns the timestamp in milliseconds."""
        return self.json["origin_server_ts"]

    @property
    def id(self):
        """Returns the event id."""
        return self.json["room_id"]


class EventStream:
    """Iterable representing a stream of events sent by the HS.

    Usage:
        events = EventStream(event_list, api)
        [print(i.user) for i in events]
        """

    def __init__(self, json, api_factory):
        """Instantiates EventStream instance.

        Args:
            json(list): List from deserializing txn from homeserver.
            api_factory(func): Generates http api when passed mxid.
        """
        self.json = json
        self._index = 0
        self.api_factory = api_factory

    def __iter__(self):
        return self

    def __next__(self):
        if self._index == len(self.json):
            self._index = 0
            raise StopIteration
        else:
            self._index += 1
            return Event(self.json[self._index - 1], self.api_factory)


class MatrixRoom:
    """Represents matrix room."""

    def __init__(self, room_id, api):
        """Instantiates MatrixRoom object.

        Args:
            room_id(str): Matrix room id (e.g. !1234567:example.com)
            api(MatrixASHttpAPI): Api for calls to the server.
        """
        self.room_id = room_id
        self.api = api

    def send_text(self, text):
        self.api.send_message(self.room_id, text)

    def send_html(self, html, body=None):
        raise NotImplementedError("Use MatrixRoom.api for now.")

    def send_emote(self, text):
        self.api.send_emote(self.room_id, text)

    def send_notice(self, text):
        self.api.send_notice(self.room_id, text)

    def send_file(self, url, file_object, **extra_information):
        raise NotImplementedError("Methods need to be implemented in api.")

    def send_location(self, body, geo_uri, thumbnail_url=None,
                      thumbnail_info={}):
        self.api.send_location(self.room_id, geo_uri, body)

    def invite(self, user_id):
        self.api.invite_user(self.room_id, user_id)

    def kick(self, user_id, reason=""):
        self.api.kick_user(self.room_id, user_id, reason=reason)

    def ban(self, user_id, reason=""):
        self.api.ban_user(self.room_id, user_id, reason=reason)

    def unban(self, user_id):
        self.api.unban_user(self.room_id, user_id)

    def leave(self):
        self.api.leave_room(self.room_id)

    @property
    def name(self):
        return self.api.get_room_name(self.room_id)["name"]

    @name.setter
    def name(self, new_name):
        self.api.set_room_name(self.room_id, new_name)

    @property
    def topic(self):
        return self.api.get_room_topic(self.room_id)["topic"]

    @topic.setter
    def topic(self, new_topic):
        self.api.set_room_topic(self.room_id, new_topic)

    @property
    def aliases(self):
        raise NotImplementedError(
            "Getting state event m.room.aliases not yet implemented upstream"
        )

    def add_alias(self, new_alias):
        self.api.set_room_alias(self.room_id, new_alias)

    def rm_alias(self, alias):
        self.api.remove_room_alias(alias)

    @property
    def members(self):
        return self.api.get_room_members(self.room_id)


class MatrixUser:
    """Represents matrix user account."""

    def __init__(self, mxid, api_factory):
        """Instantiates MatrixUser object.

        Args:
            mxid(str): User id (e.g. @me:example.createRoom)
            api(MatrixASHttpAPI): Api for calls to the server.
        """
        self.mxid = mxid
        self.user_api = api_factory(mxid)
        self.api = api_factory()
        self._rooms = {}

    def register(self):
        """Registers self.mxid with homeserver."""
        # Shouldn't send user_id param with registration
        return self.api.register(utils.mxid2localpart(self.mxid))

    def create_room(self, alias=None, is_public=False, invitees=()):
        """Calls /createRoom as self.mxid."""
        response = self.user_api.create_room(alias=alias, is_public=is_public,
                                             invitees=invitees)
        return self._mkroom(response["room_id"])

    def join(self, room_str):
        """Joins room id or alias even if it must first be created."""
        response = self.user_api.join_room(room_str)
        return self._mkroom(response["room_id"])

    def _mkroom(self, room_id):
        self._rooms[room_id] = MatrixRoom(room_id, self.user_api)
        return self._rooms[room_id]

    @property
    def displayname(self):
        """Queries the server to return current displayname."""
        return self.api.get_display_name(self.mxid)

    @displayname.setter
    def displayname(self, new_displayname):
        """PUTs new displayname to server."""
        self.user_api.set_display_name(self.mxid, new_displayname)

    @property
    def avatar_url(self):
        """Gets current avatar url from server."""
        return self.api.get_avatar_url(self.mxid)

    @avatar_url.setter
    def avatar_url(self, new_url):
        """PUTs new avatar url to server."""
        self.user_api.set_avatar_url(self.mxid, new_url)

    @property
    def rooms(self):
        """Refreshes room list if empty and returns it."""
        if not self._rooms:
            self.refresh_rooms()
        return self._rooms

    def refresh_rooms(self):
        """Calls GET /joined_rooms to refresh rooms list."""
        for room_id in self.user_api.get_joined_rooms()["joined_rooms"]:
            self._room[room_id] = MatrixRoom(room_id, self.user_api)



# Copyright 2016 Adam Beckmeyer
#
# This file is part of Gyr.
#
# Gyr is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Gyr is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Gyr.  If not, see
# <http://www.gnu.org/licenses/>.

from string import ascii_lowercase, digits
import uuid


def gen_txn_id():
    """Returns a new, unique txn id."""
    return str(uuid.uuid1())


def is_full_mxid(user_string):
    """Returns True if a string is a valid mxid."""
    if not user_string[0] == "@":
        return False
    parts = user_string[1:].split(":")
    localpart_chars = ascii_lowercase + digits + "._-="
    if not (len(parts) == 2 and all([i in localpart_chars for i in parts[0]])):
        return False
    return True


def mxid2localpart(mxid):
    """Returns the localpart of a valid mxid."""
    return mxid.strip("@").split(":")[0]



# Copyright 2016 Adam Beckmeyer
#
# This file is part of Gyr.
#
# Gyr is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Gyr is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Gyr.  If not, see
# <http://www.gnu.org/licenses/>.



# Copyright 2016 Adam Beckmeyer
#
# This file is part of Gyr.
#
# Gyr is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gyr is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along
# with Gyr.  If not, see <http://www.gnu.org/licenses/>.

from setuptools import setup, find_packages
import os.path

here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='gyr',
    author='Adam Beckmeyer',
    author_email='adam_git@thebeckmeyers.xyz',
    description='A python framework for building matrix application services',
    long_description=long_description,
    version='0.1.0',
    url='https://github.com/non-Jedi/gyr',
    license='GPLv3',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Internet :: WWW/HTTP :: WSGI',
        'Topic :: Communications :: Chat',
        'Programming Language :: Python :: 3',
    ],
    keywords='matrix chat falcon WSGI application-service',
    packages=find_packages(),
    package_data={'': ['LICENSE', 'README.md']},
    install_requires=[
        'falcon',
        'requests',
        'matrix-client>=0.0.6'
    ],
)



# Copyright 2016 Adam Beckmeyer
#
# This file is part of Gyr.
#
# Gyr is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Gyr is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Gyr.  If not, see
# <http://www.gnu.org/licenses/>.
from matrix_client.api import MatrixHttpApi, MATRIX_V2_API_PATH


class MatrixASHttpAPI(MatrixHttpApi):
    """Wraps MatrixHttpApi to allow AS identity assertion.

    Usage:
        matrix = MatrixASHttpAPI("https://matrix.org",
                                 identity="@matrix:matrix.org",
                                 token="foobar")
        response = matrix.sync()
        response = matrix.send_message("!roomid:matrix.org",
                                       "Hello!")
    """

    def __init__(self, *args, identity=None, **kwargs):
        super(MatrixASHttpAPI, self).__init__(*args, **kwargs)
        self.identity = identity

    def _send(self, *args, **kwargs):
        if "query_params" not in kwargs:
            kwargs["query_params"] = {}
        if self.identity:
            kwargs["query_params"]["user_id"] = self.identity

        return super(MatrixASHttpAPI, self)._send(*args, **kwargs)

    def register(self, username):
        """Performs /register with type: m.login.application_service

        Args:
            username(str): Username to register.
        """
        content = {
            "type": "m.login.application_service",
            "username": username,
        }
        return self._send("POST", "/register", content,
                          api_path=MATRIX_V2_API_PATH)

    def get_joined_rooms(self):
        """Performs GET /joined_rooms."""
        return self._send("GET", "/joined_rooms")



# Copyright 2016 Adam Beckmeyer
#
# This file is part of Gyr.
#
# Gyr is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gyr is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along
# with Gyr.  If not, see <http://www.gnu.org/licenses/>.

from gyr.server import Application
from gyr.matrix_objects import MatrixUser


application = Application("http://localhost:8008", "foobar")

example_mxid = "@_example_user:local"
example_user = MatrixUser(example_mxid, application.create_api)


def room_handler(room_alias):
    return False


def user_handler(user_id):
    return user_id == example_mxid


def transaction_handler(event_stream):
    for event in event_stream:

        if (event.type == "m.room.member" and
                event.json["content"] == {"membership": "invite"} and
                event.json["state_key"] == example_mxid):
            example_user.join(event.json["room_id"])

        elif (event.type == "m.room.message" and
              event.json["content"]["msgtype"] == "m.text"):
            user_room = example_user.rooms[event.json["room_id"]]
            user_room.send_notice(event.json["content"]["body"])

    return True


application.add_handlers(room_handler=room_handler,
                         transaction_handler=transaction_handler,
                         user_handler=user_handler)



# Copyright 2016 Adam Beckmeyer
#
# This file is part of Gyr.
#
# Gyr is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# Gyr is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with Gyr.  If not, see <http://www.gnu.org/licenses/>.

import falcon
import json
from . import utils
from .matrix_objects import EventStream


class Resource:
    """Master class for falcon http resources."""

    def __init__(self, handler, api_factory):
        self.handler = handler
        self.api_factory = api_factory
        # Api used directly by Room and User resources
        self.api = api_factory()
        # Place to track occurences of e.g. a certain txn_id
        self.tracker = [None] * 20

    def _is_new(self, identifier):
        """Returns True if identifier hasn't been seen before."""
        if identifier in self.tracker:
            return False
        else:
            self.tracker.append(identifier)
            self.tracker.pop(0)
            return True


class Room(Resource):
    """Generates objects to respond to GET /rooms/{room_alias}."""

    def on_get(self, request, response, room_alias=None):
        """Called when a GET request is sent to /rooms/{room_alias}"""
        response.body = "{}"
        if self.handler(room_alias):
            response.status = falcon.HTTP_200
            self.api.create_room(alias=room_alias)
        else:
            response.status = falcon.HTTP_404


class Transaction(Resource):
    """Generates objects to respond to PUT /transactions/{txn_id}"""

    def on_put(self, request, response, txn_id=None):
        """Responds to PUT request containing events."""
        response.body = "{}"

        # Check whether repeat txn_id
        if not self._is_new(txn_id):
            response.status = falcon.HTTP_200
            return

        request.context["body"] = request.stream.read()
        try:
            events = json.loads(request.context["body"].decode("utf-8"))["events"]
        except(KeyError, ValueError, UnicodeDecodeError):
            response.status = falcon.HTTP_400
            response.body = "Malformed request body"
            return

        if self.handler(EventStream(events, self.api_factory)):
            response.status = falcon.HTTP_200
        else:
            response.status = falcon.HTTP_400


class User(Resource):
    """Generates objects to respond to GET /users/{user_id}"""

    def on_get(self, request, response, user_id=None):
        """Responds to GET request for users."""
        response.body = "{}"
        if self.handler(user_id):
            response.status = falcon.HTTP_200
            self.api.register(utils.mxid2localpart(user_id))
        else:
            response.status = falcon.HTTP_404
